import os
import numpy as np
import torch
import torchmetrics
from torch import nn
from torch.nn import functional as F
import pytorch_lightning as pl
from typing import Callable, Optional, Dict, Tuple
from TSFEDL.blocks_pytorch import RTABlock, SqueezeAndExcitationModule, DenseNetDenseBlock, DenseNetTransitionBlock, \
    SpatialAttentionBlockZhangJin, TemporalAttentionBlockZhangJin
from TSFEDL.utils import flip_indices_for_conv_to_lstm, flip_indices_for_conv_to_lstm_reshape
from TSFEDL.utils import TimeDistributed



class TSFEDL_BaseModule(pl.LightningModule):
    """
    Base module for any pyTorch Lightning based algorithm in this library

    Parameters
    ----------
        in_features: int,
            The number of input features.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: List[Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            A dictionary of metrics to be returned in the test phase. These metric are functions that accepts two tensors as
            inputs, i.e., predictions and targets, and return another tensor with the metric value. Defaults contains the accuracy

        top_module: nn.Module, defaults=None
            The optional nn.Module to be used as additional top layers.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.
    """

    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = None,
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(TSFEDL_BaseModule, self).__init__()
        self.kwargs = kwargs
        self.in_features = in_features
        self.loss = loss
        self.optimizer_ = optimizer
        self.classifier = top_module
        self.metrics = metrics
        pass

    def forward(self, x):
        pass

    def training_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self(x)
        loss = self.loss(y_hat, y)
        self.log('train_loss', loss, prog_bar=True)
        #self.log('train_acc', self.accuracy(y_hat, y))
        return loss
    
    def validation_step(self, batch, batch_idx):
        # this is the validation loop
        x,y = batch
        y_hat = self(x)
        val_loss = self.loss(y_hat, y)
        self.log("val_loss", val_loss, prog_bar=True)
        return val_loss

    def test_step(self, batch, batch_idx):
        x, y = batch
        y_hat = self(x)

        # compute test loss
        loss = self.loss(y_hat, y)
        self.log('test_loss', loss, on_step=True, on_epoch=True)

        # compute the remaining test metrics
        for name, f in self.metrics.items():
            value = f(y_hat, y)
            self.log(str('test_' + name), value, on_step=True, on_epoch=True)


    def configure_optimizers(self):
        opt = self.optimizer_(self.parameters(), **self.kwargs)
        return opt


class OhShuLih_Classifier(nn.Module):
    """
    Classifier of the OhShuLi model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """

    def __init__(self, in_features, n_classes):
        super(OhShuLih_Classifier, self).__init__()
        self.model = nn.Sequential(
            nn.Dropout(p=0.2),
            nn.Linear(in_features=in_features, out_features=20),
            nn.ReLU(),
            nn.Linear(in_features=20, out_features=10),
            nn.ReLU(),
            nn.Linear(in_features=10, out_features=n_classes)
        )

    def forward(self, x):
        return self.model(x)


class OhShuLih(TSFEDL_BaseModule):
    """
    CNN+LSTM model for Arrythmia classification.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=OhShuLih_Classifier(20,5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.

    References
    ----------
        `Oh, Shu Lih, et al. "Automated diagnosis of arrhythmia using combination of CNN and LSTM techniques with
        variable length heart beats." Computers in biology and medicine 102 (2018): 278-287.`
    """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = OhShuLih_Classifier(in_features=20, n_classes=5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(OhShuLih, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        conv_layers = []

        # The padding of (kernel_size - 1) is for full convolution.
        # We use a Lazy Conv1D layer here as we dont know the input channels beforehand.
        # NOTE: INPUT SHAPE MUST BE (N, C, L) where N is the batch size, C is the NUMBER OF CHANNEL (as opposite to keras)
        # and L is the number of timesteps of Length of the 1D signal.
        conv_layers.append(nn.Conv1d(in_channels=in_features, out_channels=3, kernel_size=20, bias=False, stride=1, padding=19))
        conv_layers.append(nn.ReLU())
        conv_layers.append(nn.MaxPool1d(kernel_size=2))

        # The remaining convolutional layers can be normal ones: we know the input size.
        conv_layers.append(nn.Conv1d(in_channels=3, out_channels=6, kernel_size=10, bias=False, padding=9))
        conv_layers.append(nn.ReLU())
        conv_layers.append(nn.MaxPool1d(kernel_size=2))

        conv_layers.append(nn.Conv1d(in_channels=6, out_channels=6, kernel_size=5, bias=False, padding=4))
        conv_layers.append(nn.ReLU())
        conv_layers.append(nn.MaxPool1d(kernel_size=2))

        self.convolutions = nn.Sequential(*conv_layers)

        # input_size is the number of DIMENSIONS of the data.
        # From the convolution above, the number of output channels is 6, so we have 6 dimensions in our data.
        # NOTE: In keras, this LSTM layer employs recurrent dropout. THIS IS NOT IMPLEMENTED IN PYTORCH!!
        # NOTE2: this LSTM Requires an input with shape (N, L, H_in) where N is the batch size, L is the LENGTH of the sequence
        # and H_in is the number of dimensions. From the convolutional layers we got a shape of (N, H_in, L), so
        # we have to RESHAPE our BEFORE plugging them into the LSTM.
        self.lstm = nn.LSTM(input_size=6, hidden_size=20, batch_first=True)


    def forward(self, x):
        out = self.convolutions(x)
        # Now, flip indices using a view for the LSTM as it requires a shape of (N, L, H_in = C)
        out = out.view(out.size(0), out.size(2), out.size(1))
        out, _ = self.lstm(out)

        if self.classifier is not None:
            out = self.classifier(out[:, -1, :])  # We only want the last step of the LSTM output

        return out


def en_loss(y_pred: torch.Tensor, y_true: torch.Tensor) -> torch.Tensor:
    """
    YiboGao's custom loss function
    """
    epsilon = 1.e-7
    gamma = float(0.3)

    y_pred = torch.clip(y_pred, epsilon, 1.0 - epsilon)
    pos_pred = torch.pow(- torch.log(y_pred), gamma)
    neg_pred = torch.pow(- torch.log(1 - y_pred), gamma)
    y_t = torch.multiply(y_true, pos_pred) + torch.multiply(1 - y_true, neg_pred)
    loss = torch.mean(y_t)

    return loss


class YiboGaoClassifier(nn.Module):
    """
    Classifier of the Yibo Gao model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes):
        super(YiboGaoClassifier, self).__init__()
        self.n_classes = n_classes
        self.in_features = in_features
        self.module = nn.Sequential(
            nn.Flatten(),
            nn.Dropout(p=0.7),
            nn.Linear(in_features=in_features, out_features=100),
            nn.ReLU(),
            nn.Dropout(p=0.7),
            nn.Linear(100, n_classes),
            nn.Softmax(dim=1)
        )

    def forward(self, x):
        return self.module(x)


class YiboGao(TSFEDL_BaseModule):
    """
    CNN using RTA blocks for end-to-end attrial fibrilation detection.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=YiboGaoClassifier(384, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Gao, Y., Wang, H., & Liu, Z. (2021). An end-to-end atrial fibrillation detection by a novel residual-based
        temporal attention convolutional neural network with exponential nonlinearity loss.
        Knowledge-Based Systems, 212, 106589.

    Notes
    -----
        Code adapted from the original implementation available at:
            https://github.com/o00O00o/RTA-CNN
    """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = YiboGaoClassifier(384, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = en_loss,
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(YiboGao, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        # Model definition
        self.module = nn.Sequential(
            RTABlock(in_features, 16, 32),
            nn.MaxPool1d(kernel_size=4),
            RTABlock(16, 32, 16),
            nn.MaxPool1d(kernel_size=4),
            RTABlock(32, 64, 9),
            nn.MaxPool1d(kernel_size=2),
            RTABlock(64, 64, 9),
            nn.MaxPool1d(kernel_size=2),
            nn.Dropout(p=0.6),
            RTABlock(64, 128, 3),
            nn.MaxPool1d(kernel_size=2),
            RTABlock(128, 128, 3),
            nn.MaxPool1d(kernel_size=2),
            nn.Dropout(p=0.6)
        )

    def forward(self, x):
        x = self.module(x)

        if self.classifier is not None:
            x = self.classifier(x)

        return x


class YaoQihangClassifier(nn.Module):
    """
    Classifier of the Yao Qihang model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes):
        super(YaoQihangClassifier, self).__init__()
        self.module = nn.Sequential(
            nn.Linear(in_features=in_features, out_features=32),
            nn.Tanh(),
            nn.Linear(32, n_classes),
            # On each timestep it makes the softmax. Then, the final classification probablity is the average
            # throughout all the timesteps
            # nn.Softmax(dim=2)
        )

    def forward(self, x):
        x1 = self.module(x)
        # NOTE:  This classifier must return the mean softmax over ALL TIMESTEPS. TAKE CARE OF THE LOSS FUNCTION!
        mean = torch.mean(F.softmax(x1, dim=2), dim=1)
        return mean


class YaoQihang(TSFEDL_BaseModule):
    """
    Attention-based time-incremental convolutional neural network (ATI-CNN)

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=YaoQihangClassifier(32, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Yao, Q., Wang, R., Fan, X., Liu, J., & Li, Y. (2020). Multi-class Arrhythmia detection from 12-lead varied-length
        ECG using Attention-based Time-Incremental Convolutional Neural Network. Information Fusion, 53, 174-182.

    """
    def __convBlock(self, in_features, filters) -> nn.Module:
        return nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=filters, kernel_size=3, padding='same'),
            nn.BatchNorm1d(num_features=filters),
            nn.ReLU()
        )

    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = YaoQihangClassifier(32, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = None,
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs):
        super(YaoQihang, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            self.__convBlock(in_features, 64),
            self.__convBlock(64, 64),
            nn.MaxPool1d(kernel_size=3, stride=3),
            self.__convBlock(64, 128),
            self.__convBlock(128, 128),
            nn.MaxPool1d(kernel_size=3, stride=3),
            self.__convBlock(128, 256),
            self.__convBlock(256, 256),
            self.__convBlock(256, 256),
            nn.MaxPool1d(kernel_size=3, stride=3),
            self.__convBlock(256, 256),
            self.__convBlock(256, 256),
            self.__convBlock(256, 256),
            nn.MaxPool1d(kernel_size=3, stride=3),
            self.__convBlock(256, 256),
            self.__convBlock(256, 256),
            self.__convBlock(256, 256),
            nn.MaxPool1d(kernel_size=3, stride=3),
        )

        # Temporal Layers (2 stacked LSTM): REMEMBER TO SWAP DIMENSIONS ON THE FORWARD METHOD.
        # input is 256 as we expect the last convolution with 256 filters.
        self.lstm = nn.LSTM(input_size=256, hidden_size=32, num_layers=2, dropout=0.2, batch_first=True)


    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm(x)  # We don't care about hidden state.

        if self.classifier is not None:
            x = self.classifier(x)

        return x


class HtetMyetLynn(TSFEDL_BaseModule):
    """
    Hybrid CNN+Bidirectional RNN

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        use_rnn: str, ["gru", "lstm", None], defaults="gru"
            This parameter controls wether to use RNN layers and what type to use

        top_module: nn.Module
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Lynn, H. M., Pan, S. B., & Kim, P. (2019). A deep bidirectional GRU network model for biometric
        electrocardiogram classification based on recurrent neural networks. IEEE Access, 7, 145395-145405.
    """
    def __init__(self,
                 in_features: int,
                 use_rnn: Optional[str] = 'gru',  # Options are 'gru' or 'lstm' or None
                 top_module: Optional[nn.Module] = nn.Sequential(nn.Linear(in_features=80, out_features=5)),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(HtetMyetLynn, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convLayers = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=30, kernel_size=5, padding='same'),
            nn.MaxPool1d(kernel_size=2),
            nn.Conv1d(in_channels=30, out_channels=30, kernel_size=2, padding='same'),
            nn.MaxPool1d(kernel_size=2),
            nn.Conv1d(in_channels=30, out_channels=60, kernel_size=5, padding='same'),
            nn.MaxPool1d(kernel_size=2),
            nn.Conv1d(in_channels=60, out_channels=60, kernel_size=2, padding='same'),
            nn.MaxPool1d(kernel_size=2),
        )

        if use_rnn is not None:
            if use_rnn.lower() == 'gru':
                self.rnn = nn.GRU(input_size=60, hidden_size=40, dropout=0.2, bidirectional=True, batch_first=True)
            else:
                self.rnn = nn.LSTM(input_size=60, hidden_size=40, dropout=0.2, bidirectional=True, batch_first=True)
        else:
            self.rnn = None

    def forward(self, x):
        x = self.convLayers(x)

        if self.rnn is not None:
            x = flip_indices_for_conv_to_lstm(x)
            x, _ = self.rnn(x)

        if self.classifier is not None:
            x = self.classifier(x[:, -1, :])  # get only the last timestep

        return x


class YildirimOzal(TSFEDL_BaseModule):
    """CAE-LSTM

    Parameters
    ----------
        input_shape: Tuple[int, int]
            Shape of the input tensors (number of channels, sequence length)

        train_autoencoder: bool, defaults=True
            This parameter controls wether to train the autoencoder or the classifier

        top_module: nn.Module
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        autoenconder: `LightningModule`
            A pyTorch Lightning Module instance with the autoencoder.

        encoder: `LightningModule`
            A pyTorch Lightning Module instance with the encoder.

        model: `LightningModule`
            A pyTorch Lightning Module instance with the classifier.

    References
    ----------
        Yildirim, Ozal, et al. "A new approach for arrhythmia classification using deep coded features and LSTM
        networks." Computer methods and programs in biomedicine 176 (2019): 121-133.
     """
    def __init__(self,
                 input_shape: Tuple[int, int],    # (Channels, seq_length)
                 train_autoencoder: bool = True,  # This is for training the autoencoder or the LSTM-classifier
                 top_module: Optional[nn.Module] = None,
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(YildirimOzal, self).__init__(input_shape[0], top_module, loss, metrics, optimizer, **kwargs)
        self.train_autoencoder = train_autoencoder

        self.encoder = nn.Sequential(
            nn.Conv1d(in_channels=self.in_features, out_channels=16, kernel_size=5, padding='same'),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2),
            nn.Conv1d(in_channels=16, out_channels=64, kernel_size=5, padding='same'),
            nn.ReLU(),
            nn.BatchNorm1d(num_features=64),
            nn.MaxPool1d(kernel_size=2),
            nn.Conv1d(in_channels=64, out_channels=32, kernel_size=3, padding='same'),
            nn.ReLU(),
            nn.Conv1d(in_channels=32, out_channels=1, kernel_size=3, padding='same'),
            nn.MaxPool1d(kernel_size=2)
        )
        output_length = input_shape[1] // 2 // 2 // 2  # seq_length is reduced by a half three times (3 maxpooling)

        self.decoder = nn.Sequential(
            nn.Conv1d(in_channels=1, out_channels=1, kernel_size=3, padding='same'),
            nn.ReLU(),
            nn.Conv1d(in_channels=1, out_channels=32, kernel_size=3, padding='same'),
            nn.ReLU(),
            nn.Upsample(scale_factor=2),
            nn.Conv1d(in_channels=32, out_channels=64, kernel_size=5, padding='same'),
            nn.ReLU(),
            nn.Upsample(scale_factor=2),
            nn.Conv1d(in_channels=64, out_channels=16, kernel_size=5, padding='same'),
            nn.ReLU(),
            nn.Flatten(),
            nn.Linear(16 * output_length * 2 * 2, input_shape[1])  # Reconstruction (16 -> out_channels for conv, and 2 upsamples.
        )

        self.lstm = nn.LSTM(input_size=1, hidden_size=32, batch_first=True)

    def forward(self, x):
        reduction = self.encoder(x)
        if self.train_autoencoder:
            reconstruction = self.decoder(reduction)
            reconstruction = reconstruction.view(reconstruction.size(0), 1, reconstruction.size(1))
            return reconstruction
        else:
            x = flip_indices_for_conv_to_lstm(x)
            x, _ = self.lstm(x)
            if self.classifier is not None:
                x = self.classifier(x[:, -1, :])
            return x

    def training_step(self, batch, batch_idx):
        if self.train_autoencoder:
            x, _ = batch
            x_hat = self(x)
            loss = self.loss(x_hat, x)
            self.log('train_loss', loss)
            return loss
        else:
            x, y = batch
            y_hat = self(x)
            loss = self.loss(y_hat, y)
            self.log('train_loss', loss)
            return loss


class CaiWenjuan(TSFEDL_BaseModule):  # TODO: Squeeze and Activation units depends on input size.
    """
    DDNN network presented in:

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        reduction_ratio: float, defaults=0.6
            This parameter controls the reduction of the dense transition block

        top_module: nn.Module
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        `Cai, Wenjuan, et al. "Accurate detection of atrial fibrillation from 12-lead ECG using deep neural network."
        Computers in biology and medicine 116 (2020): 103378.`
   """
    def __init__(self,
                 in_features: int,
                 reduction_ratio: float = 0.6,
                 top_module: Optional[nn.Module] = nn.Sequential(nn.Linear(67, 5)),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(CaiWenjuan, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        # Convolutional inputs
        self.conv1 = nn.Conv1d(in_channels=in_features, out_channels=8, kernel_size=1, padding='same')
        self.conv2 = nn.Conv1d(in_channels=in_features, out_channels=16, kernel_size=3, padding='same')
        self.conv3 = nn.Conv1d(in_channels=in_features, out_channels=24, kernel_size=5, padding='same')
        # NOTE: Concatenate layers after that (cat or stack)
        self.in_features_after_concat = 8 + 16 + 24

        # dense-block construction
        dense_layers = [2, 4, 6, 4]
        in_feat = self.in_features_after_concat
        layers = []
        for i, n_block in enumerate(dense_layers):
            layers.append(SqueezeAndExcitationModule(in_feat, 32))
            layers.append(DenseNetDenseBlock(in_features=in_feat, layers=n_block, growth_rate=6))
            in_feat = layers[-1].output_features
            layers.append(SqueezeAndExcitationModule(in_feat, 32))
            if i < len(dense_layers) - 1:
                layers.append(DenseNetTransitionBlock(in_features=in_feat, reduction=reduction_ratio))
                in_feat = layers[-1].out_features

        self.dense_module = nn.Sequential(*layers)
        self.output_features = in_feat

    def forward(self, x):
        x1 = self.conv1(x)
        x2 = self.conv2(x)
        x3 = self.conv3(x)

        x = torch.cat((x1, x2, x3), dim=1)
        x = self.dense_module(x)

        # Global average pooling (average value for each channel
        x = torch.mean(x, dim=2)

        if self.classifier is not None:
            x = self.classifier(x)

        return x


class ZhangJin_Classifier(nn.Module):
    """
    Classifier of the Zhang Jin model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes):
        super(ZhangJin_Classifier, self).__init__()
        self.linear = nn.Sequential(
            nn.Linear(in_features, n_classes)
        )

    def forward(self, x):
        return self.linear(x[:, -1, :])


class ZhangJin(TSFEDL_BaseModule):
    """
    A CNN+Bi-Directional GRU with a spatio-temporal attention mechanism.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        decrease_ratio: int, defaults=2
            This parameter controls the decrease ratio of the spatial attention block

        top_module: nn.Module, defaults=ZhangJin_Classifier(24, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Zhang, J., Liu, A., Gao, M., Chen, X., Zhang, X., & Chen, X. (2020). ECG-based multi-class arrhythmia detection
        using spatio-temporal attention-based convolutional recurrent neural network.
        Artificial Intelligence in Medicine, 106, 101856.
   """
    def __convBlock(self, in_features, filters) -> nn.Module:
        return nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=filters, kernel_size=3, padding='same'),
            nn.BatchNorm1d(num_features=filters),
            nn.ReLU()
        )

    def __init__(self,
                 in_features: int,
                 decrease_ratio: int = 2,
                 top_module: Optional[nn.Module] = ZhangJin_Classifier(24, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(ZhangJin, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)
        self.decrease_ratio = decrease_ratio

        self.convolutions = nn.ModuleList(
            [
                nn.Sequential(
                    self.__convBlock(in_features, 64),
                    self.__convBlock(64, 64),
                    nn.MaxPool1d(kernel_size=3, stride=3),
                    nn.Dropout(p=0.2)
                ),
                nn.Sequential(      # Features after temporal attention is 1.
                    self.__convBlock(64, 128),
                    self.__convBlock(128, 128),
                    nn.MaxPool1d(kernel_size=3, stride=3),
                    nn.Dropout(p=0.2)
                ),
                nn.Sequential(
                    self.__convBlock(128, 256),
                    self.__convBlock(256, 256),
                    self.__convBlock(256, 256),
                    nn.MaxPool1d(kernel_size=3, stride=3),
                    nn.Dropout(p=0.2)
                ),
                nn.Sequential(
                    self.__convBlock(256, 256),
                    self.__convBlock(256, 256),
                    self.__convBlock(256, 256),
                    nn.MaxPool1d(kernel_size=3, stride=3),
                    nn.Dropout(p=0.2),
                ),
                nn.Sequential(
                    self.__convBlock(256, 256),
                    self.__convBlock(256, 256),
                    self.__convBlock(256, 256),
                    nn.MaxPool1d(kernel_size=3, stride=3),
                    nn.Dropout(p=0.2)
                )
            ]
        )
        self.spatial_attention = nn.ModuleList(
            [SpatialAttentionBlockZhangJin(f, decrease_ratio) for f in [64, 128, 256, 256, 256]]
        )
        self.temporal_attention = nn.ModuleList(
            [TemporalAttentionBlockZhangJin() for i in range(5)]
        )

        self.gru = nn.GRU(input_size=256, hidden_size=12, batch_first=True, bidirectional=True, dropout=0.2)

    def forward(self, x):
        for i in range(5):
            x = self.convolutions[i](x)
            x_spatial = self.spatial_attention[i](x)
            x = torch.multiply(x, x_spatial)
            x_temp = self.temporal_attention[i](x)
            x = torch.multiply(x, x_temp)

        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.gru(x)
        if self.classifier is not None:
            x = self.classifier(x)
        return x


class KongZhengmin_Classifier(nn.Module):
    """
    Classifier of the Kong Zhengmin model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence=False):
        super(KongZhengmin_Classifier, self).__init__()
        self.return_sequnce = return_sequence
        self.module = nn.Sequential(
            nn.Linear(in_features=in_features, out_features=50),
            nn.ReLU(),
            nn.Linear(in_features=50, out_features=50),
            nn.ReLU(),
            nn.Linear(in_features=50, out_features=n_classes)
        )

    def forward(self, x):
        if self.return_sequnce:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class KongZhengmin(TSFEDL_BaseModule):
    """
    CNN+LSTM

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=KongZhengmin_Classifier(64, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Kong, Zhengmin, et al. "Convolution and Long Short-Term Memory Hybrid Deep Neural Networks for Remaining
        Useful Life Prognostics." Applied Sciences 9.19 (2019): 4156.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = KongZhengmin_Classifier(64, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(KongZhengmin, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolution = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=32, kernel_size=5),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2)
        )

        self.lstm = nn.LSTM(batch_first=True, input_size=32, hidden_size=64, num_layers=2)

    def forward(self, x):
        x = self.convolution(x)

        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x


class WeiXiaoyan_Classifier(nn.Module):
    """
    Classifier of the Wei Xiaoyan model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes):
        super(WeiXiaoyan_Classifier, self).__init__()
        self.module = nn.Sequential(
            nn.Linear(in_features, 1024),
            nn.ReLU(),
            nn.Linear(1024, n_classes),
        )

    def forward(self, x):
        return self.module(x)


class WeiXiaoyan(TSFEDL_BaseModule):
    """
    CNN+LSTM model employed for 2-D images of ECG data. This model is adapted for 1-D time series.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=WeiXiaoyan_Classifier(512, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Wei, Xiaoyan, et al. "Early prediction of epileptic seizures using a long-term recurrent convolutional
        network." Journal of neuroscience methods 327 (2019): 108395.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = WeiXiaoyan_Classifier(512, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(WeiXiaoyan, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=32, kernel_size=5),
            nn.LeakyReLU(negative_slope=0.01),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.BatchNorm1d(num_features=32),
            nn.Conv1d(in_channels=32, out_channels=64, kernel_size=3),
            nn.LeakyReLU(negative_slope=0.01),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.BatchNorm1d(num_features=64),
            nn.Conv1d(in_channels=64, out_channels=128, kernel_size=3),
            nn.LeakyReLU(negative_slope=0.01),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.BatchNorm1d(num_features=128),
            nn.Conv1d(in_channels=128, out_channels=256, kernel_size=3),
            nn.LeakyReLU(negative_slope=0.01),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.BatchNorm1d(num_features=256),
            nn.Conv1d(in_channels=256, out_channels=512, kernel_size=3),
            nn.LeakyReLU(negative_slope=0.01),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.BatchNorm1d(num_features=512),
        )
        self.lstm1 = nn.LSTM(input_size=512, hidden_size=512, batch_first=True)
        self.batchNorm = nn.BatchNorm1d(num_features=512)
        self.lstm2 = nn.LSTM(input_size=512, hidden_size=512, batch_first=True)

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm1(x)
        x = x.reshape((x.size(0), x.size(2), x.size(1)))
        x = self.batchNorm(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm2(x)

        if self.classifier is not None:
            x = self.classifier(x[:, -1, :])
        return x


class GaoJunLi_Classifier(nn.Module):
    """
    Classifier of the Gao Jun Li model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence=False):
        super(GaoJunLi_Classifier, self).__init__()

        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Linear(in_features, 32),
            nn.ReLU(),
            nn.Linear(32, n_classes)
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:,-1,:])


class GaoJunLi(TSFEDL_BaseModule):
    """
    LSTM network

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=GaoJunLi_Classifier(64, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Gao, Junli, et al. "An Effective LSTM Recurrent Network to Detect Arrhythmia on Imbalanced ECG Dataset."
        Journal of healthcare engineering 2019 (2019).
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = GaoJunLi_Classifier(64, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(GaoJunLi, self).__init__(in_features, top_module, loss, optimizer, **kwargs)
        self.lstm = nn.LSTM(input_size=in_features, hidden_size=64, dropout=0.3)

    def forward(self, x):
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm(x)
        if self.classifier is not None:
            x = self.classifier(x)
        return x


class LiOhShu_Classifier(nn.Module):
    """
    Classifier of the Li Oh Shu model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(LiOhShu_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Linear(in_features, 8),
            nn.ReLU(),
            nn.Dropout(p=0.5),
            nn.Linear(8, 8),
            nn.ReLU(),
            nn.Dropout(p=0.5),
            nn.Linear(8, n_classes)
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class LihOhShu(TSFEDL_BaseModule):
    """
    CNN+LSTM model

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=LiOhShu_Classifier(10, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Lih, Oh Shu, et al. "Comprehensive electrocardiographic diagnosis based on deep learning." Artificial
        Intelligence in Medicine 103 (2020): 101789.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = LiOhShu_Classifier(10, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(LihOhShu, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=3, kernel_size=20, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=3, out_channels=6, kernel_size=10, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=6, out_channels=6, kernel_size=5, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=6, out_channels=6, kernel_size=5, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=6, out_channels=6, kernel_size=10, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2)
        )

        self.lstm = nn.LSTM(input_size=6, hidden_size=10, batch_first=True)

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x

#####################################################################

class KhanZulfiqar_Classifier(nn.Module):
    """
    Classifier of the Khan Zulfiqar model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(KhanZulfiqar_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Linear(in_features, 8),
            nn.ReLU(),
            nn.Dropout(p=0.5),
            nn.Linear(8, 8),
            nn.ReLU(),
            nn.Dropout(p=0.5),
            nn.Linear(8, n_classes)
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class KhanZulfiqar(TSFEDL_BaseModule):
    """
    CNN+GRU model for electricity forecasting

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=KhanZulfiqar_Classifier(10, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Sajjad, M., Khan, Z. A., Ullah, A., Hussain, T., Ullah, W., Lee, M. Y., & Baik, S. W. (2020). A novel
        CNN-GRU-based hybrid approach for short-term residential load forecasting. IEEE Access, 8, 143759-143768.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = KhanZulfiqar_Classifier(10, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(KhanZulfiqar, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=3, kernel_size=20, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=3, out_channels=6, kernel_size=10, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=6, out_channels=6, kernel_size=5, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=6, out_channels=6, kernel_size=5, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=6, out_channels=6, kernel_size=10, bias=False),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2)
        )

        self.lstm = nn.LSTM(input_size=6, hidden_size=10, batch_first=True)

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x

class ZhengZhenyu_Classifier(nn.Module):
    """
    Classifier of the Zheng Zhenyu model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(ZhengZhenyu_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Linear(in_features, 2048),
            nn.ELU(),
            nn.Dropout(p=0.5),
            nn.Linear(2048, 2048),
            nn.ELU(),
            nn.Linear(2048, n_classes)
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class ZhengZhenyu(TSFEDL_BaseModule):
    """
    CNN+LSTM network for arrythmia detection. This model initialy was designed to deal with 2D images. It was adapted
    to 1D time series.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=ZhengZhenyu_Classifier(256, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Zheng, Z., Chen, Z., Hu, F., Zhu, J., Tang, Q., & Liang, Y. (2020). An automatic diagnosis of arrhythmias using
        a combination of CNN and LSTM technology. Electronics, 9(1), 121.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = ZhengZhenyu_Classifier(256, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(ZhengZhenyu, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=64, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.BatchNorm1d(num_features=64),
            nn.Conv1d(in_channels=64, out_channels=64, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=64, out_channels=128, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.BatchNorm1d(num_features=128),
            nn.Conv1d(in_channels=128, out_channels=128, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=128, out_channels=256, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.BatchNorm1d(num_features=256),
            nn.Conv1d(in_channels=256, out_channels=256, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.MaxPool1d(kernel_size=2, stride=2)
        )

        self.lstm = nn.LSTM(input_size=256, hidden_size=256, batch_first=True)

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x

#
# class HouBoroui(TSFEDL_BaseModule):
#     def __init__(self,
#                  in_features: int,
#                  encoder_units: int = 100,
#                  top_module: Optional[nn.Module] = None,
#                  loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
#                  metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
#                  optimizer: torch.optim.Optimizer = torch.optim.Adam,
#                  **kwargs
#                  ):
#         super(HouBoroui, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)
#
#         self.lstm1 = nn.LSTM(input_size=in_features, hidden_size=encoder_units, batch_first=True)
#         self.lstm2 = nn.LSTM(input_size=encoder_units, hidden_size=encoder_units, batch_first=True)
#         self.td = TimeDistributed(nn.Sequential(nn.Linear(encoder_units,1)))
#
#         # self.encoder = nn.Sequential(
#         #     nn.LSTM(input_size=in_features, hidden_size=encoder_units, batch_first=True),
#         #     nn.LSTM(input_size=encoder_units, hidden_size=encoder_units, batch_first=True),
#         #     TimeDistributed(nn.Sequential(nn.Linear(encoder_units,1)))
#         # )
#
#     def forward(self, x):
#         x = flip_indices_for_conv_to_lstm(x)
#         x, _ = self.lstm1(x)
#         x, _ = self.lstm2(x)
#         x = self.td(x)
#
#         if self.classifier is not None:
#             x = self.classifier(x)
#
#         return x

class WangKejun_Classifier(nn.Module):
    """
    Classifier of the Wang Kejun model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(WangKejun_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Flatten(),
            nn.Linear(in_features, 2048),
            nn.ELU(),
            nn.Dropout(p=0.5),
            nn.Linear(2048, 2048),
            nn.ELU(),
            nn.Linear(2048, n_classes)
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class WangKejun(TSFEDL_BaseModule):
    """
    CNN 1D-LSTM

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=WangKejun_Classifier(256, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Wang, Kejun, Xiaoxia Qi, and Hongda Liu. "Photovoltaic power forecasting based LSTM-Convolutional Network."
        Energy 189 (2019): 116225.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = WangKejun_Classifier(256, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(WangKejun, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=64, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.BatchNorm1d(num_features=64),
            nn.Conv1d(in_channels=64, out_channels=64, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=64, out_channels=128, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.BatchNorm1d(num_features=128),
            nn.Conv1d(in_channels=128, out_channels=128, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=128, out_channels=256, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.BatchNorm1d(num_features=256),
            nn.Conv1d(in_channels=256, out_channels=256, kernel_size=3, stride=1, bias=True, padding="same"),
            nn.ELU(),
            nn.MaxPool1d(kernel_size=2, stride=2)
        )

        self.lstm = nn.LSTM(input_size=256, hidden_size=256, batch_first=True)

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x

class ChenChen_Classifier(nn.Module):
    """
    Classifier of the Chen Chen model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(ChenChen_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Flatten(),
            nn.Linear(in_features, n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class ChenChen(TSFEDL_BaseModule):
    """
    CNN+LSTM model for arrythmia classification.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=ChenChen_Classifier(64, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Chen, Chen, et al. "Automated arrhythmia classification based on a combination network of CNN and LSTM."
        Biomedical Signal Processing and Control 57 (2020): 101819.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = ChenChen_Classifier(64, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(ChenChen, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=251, kernel_size=5, stride=1, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=251, out_channels=150, kernel_size=5, stride=1, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=150, out_channels=100, kernel_size=10, stride=1, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=100, out_channels=81, kernel_size=20, stride=1, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=81, out_channels=61, kernel_size=20, stride=1, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
            nn.Conv1d(in_channels=61, out_channels=14, kernel_size=10, stride=1, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2),
        )

        self.lstm1 = nn.LSTM(input_size=14, hidden_size=32, batch_first=True)
        self.lstm2 = nn.LSTM(input_size=32, hidden_size=64, batch_first=True)

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm1(x)
        x, _ = self.lstm2(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x


class KimTaeYoung_Classifier(nn.Module):
    """
    Classifier of the Kim Tae Young model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(KimTaeYoung_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Linear(in_features, 32),
            nn.ReLU(),
            nn.Linear(32, n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class KimTaeYoung(TSFEDL_BaseModule):
    """
    CNN+LSTM model for arrythmia classification.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=KimTaeYoung_Classifier(64, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Kim, Tae-Young, and Sung-Bae Cho. "Predicting residential energy consumption using CNN-LSTM neural networks."
        Energy 182 (2019): 72-81.

    Notes
    -----
        In the original paper, the time-series is windowed. Therefore, a TimeDistributed layer is employed before the
        LSTM to traverse all the generated windows. Here, we do not implement this as it is problem-specific.
        Please note that if you need this layer then you should add it manually.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = KimTaeYoung_Classifier(64, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(KimTaeYoung, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=64, kernel_size=2, stride=1, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=1),
            nn.Conv1d(in_channels=64, out_channels=64, kernel_size=2, stride=1, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=1)
        )

        self.lstm = nn.LSTM(input_size=64, hidden_size=64, batch_first=True)

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm(x)
        x, _ = self.lstm(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x


class GenMinxing_Classifier(nn.Module):
    """
    Classifier of the Gen Minxing model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(GenMinxing_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Flatten(),
            nn.Linear(in_features, n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class GenMinxing(TSFEDL_BaseModule):
    """
    CNN+LSTM model for arrythmia classification.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=GenMinxing_Classifier(80, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Geng, Minxing, et al. "Epileptic Seizure Detection Based on Stockwell Transform and Bidirectional Long
        Short-Term Memory." IEEE Transactions on Neural Systems and Rehabilitation Engineering 28.3 (2020): 573-580.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = GenMinxing_Classifier(80, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(GenMinxing, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.lstm = nn.LSTM(input_size=in_features, hidden_size=40, batch_first=True, bidirectional=True)

    def forward(self, x):
        x, _ = self.lstm(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x


class FuJiangmeng_Classifier(nn.Module):
    """
    Classifier of the Fu Jiangmeng model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(FuJiangmeng_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Linear(in_features, 128),
            nn.Tanh(),
            nn.Dropout(p=0.3),
            nn.Linear(128, n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class FuJiangmeng(TSFEDL_BaseModule):
    """
    CNN+LSTM model for arrythmia classification.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=FuJiangmeng_Classifier(256, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Fu, Jiangmeng, et al. "A hybrid CNN-LSTM model based actuator fault diagnosis for six-rotor UAVs." 2019
        Chinese Control And Decision Conference (CCDC). IEEE, 2019.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = FuJiangmeng_Classifier(256, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(FuJiangmeng, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=32, kernel_size=1, stride=1, bias=True, padding="same"),
            nn.Tanh(),
            nn.MaxPool1d(kernel_size=2, stride=1),
        )

        self.lstm = nn.LSTM(input_size=32, hidden_size=256, batch_first=True, dropout=0.3)

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm_reshape(x)
        x, _ = self.lstm(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x


class ShiHaotian_Classifier(nn.Module):
    """
    Classifier of the Shi Haotian model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(ShiHaotian_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module1 = nn.Sequential(
            nn.Flatten(),
            nn.Linear(in_features, 518),
            nn.ReLU(),
            nn.Linear(518, 88),
            nn.ReLU()
        )
        self.module2 = nn.Sequential(
            nn.Linear(88+32, n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        if self.return_sequence:
            x1 = self.module1(x)
            x_conc = torch.cat((x1, x[:,-1,:]), 1)
            return self.module2(x_conc)
        else:
            x1 = self.module1(x)
            x_conc = torch.cat((x1, x[:,-1,:]), 1)
            return self.module2(x_conc)


class ShiHaotian(TSFEDL_BaseModule):
    """
    CNN+LSTM model for arrythmia classification.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=None
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Shi, Haotian, et al. "Automated heartbeat classification based on deep neural network with multiple input
        layers." Knowledge-Based Systems 188 (2020): 105036.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = None,
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(ShiHaotian, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions1 = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=32, kernel_size=13, stride=2, bias=True, padding="valid"),
            nn.LeakyReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2)
        )

        self.convolutions2 = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=32, kernel_size=13, stride=1, bias=True, padding="valid"),
            nn.LeakyReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2)
        )

        self.convolutions3 = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=32, kernel_size=13, stride=2, bias=True, padding="valid"),
            nn.LeakyReLU(),
            nn.MaxPool1d(kernel_size=2, stride=2)
        )

        self.lstm = nn.LSTM(input_size=32*3, hidden_size=32, batch_first=True)

    def forward(self, x):
        x1 = self.convolutions1(x)
        x2 = self.convolutions1(x)
        x3 = self.convolutions1(x)
        x_conc = torch.cat((x1, x2, x3), 1)
        x_conc = flip_indices_for_conv_to_lstm(x_conc)
        x, _ = self.lstm(x_conc)

        if self.classifier is not None:
            x = self.classifier(x)
        return x


class HuangMeiLing_Classifier(nn.Module):
    """
    Classifier of the Huang Mei Ling model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(HuangMeiLing_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Flatten(),
            nn.Dropout(p=0.5),
            nn.Linear(in_features, 512),
            nn.ReLU(),
            nn.Linear(512, 256),
            nn.ReLU(),
            nn.Linear(256, n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class HuangMeiLing(TSFEDL_BaseModule):
    """
     CNN model, employed for 2-D images of ECG data. This model is adapted for 1-D time series

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=None
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Huang, Mei-Ling, and Yan-Sheng Wu. "Classification of atrial fibrillation and normal sinus rhythm based on
        convolutional neural network." Biomedical Engineering Letters (2020): 1-11.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = HuangMeiLing_Classifier(81, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(HuangMeiLing, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.ConstantPad1d(padding=1, value=0),
            nn.Conv1d(in_channels=in_features, out_channels=48, kernel_size=15, stride=6, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=1),
            nn.ConstantPad1d(padding=3, value=0),
            nn.Conv1d(in_channels=48, out_channels=256, kernel_size=7, stride=2, bias=True, padding="valid"),
            nn.ReLU(),
            nn.MaxPool1d(kernel_size=2, stride=1),
        )

    def forward(self, x):
        x = self.convolutions(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x

class HongTan_Classifier(nn.Module):
    """
    Classifier of the HongTan model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """

    def __init__(self, in_features, n_classes):
        super(HongTan_Classifier, self).__init__()
        self.model = nn.Sequential(
            nn.Linear(in_features=in_features, out_features=n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        return self.model(x)


class HongTan(TSFEDL_BaseModule):
    """
    Application of stacked convolutional and long short-term memory network
    for accurate identification of CAD ECG signals

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=HongTan_Classifier(4,5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.

    References
    ----------
        TAN, Jen Hong, et al. Application of stacked convolutional and long short-term memory network for accurate
        identification of CAD ECG signals. Computers in biology and medicine, 2018, vol. 94, p. 19-26.
    """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = HongTan_Classifier(in_features=4, n_classes=5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(HongTan, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        conv_layers = []

        conv_layers.append(nn.Conv1d(in_channels=in_features, out_channels=40, kernel_size=5, bias=False, stride=1))
        conv_layers.append(nn.ReLU())
        conv_layers.append(nn.MaxPool1d(kernel_size=2, stride=2))

        # The remaining convolutional layers can be normal ones: we know the input size.
        conv_layers.append(nn.Conv1d(in_channels=40, out_channels=32, kernel_size=3, bias=False, stride=1))
        conv_layers.append(nn.ReLU())
        conv_layers.append(nn.MaxPool1d(kernel_size=2, stride=2))

        self.convolutions = nn.Sequential(*conv_layers)

        self.lstm1 = nn.LSTM(input_size=32, hidden_size=32, batch_first=True)
        self.lstm2 = nn.LSTM(input_size=32, hidden_size=16, batch_first=True)
        self.lstm3 = nn.LSTM(input_size=16, hidden_size=4, batch_first=True)


    def forward(self, x):
        out = self.convolutions(x)
        # Now, flip indices using a view for the LSTM as it requires a shape of (N, L, H_in = C)
        out = out.view(out.size(0), out.size(2), out.size(1))
        out, _ = self.lstm1(out, batch_first=True)
        out, _ = self.lstm2(out, batch_first=True)
        out, _ = self.lstm3(out, batch_first=True)

        if self.classifier is not None:
            out = self.classifier(out[:, -1, :])  # We only want the last step of the LSTM output

        return out


class SharPar_Classifier(nn.Module):
    """
    Classifier of the SharPar model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(SharPar_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Linear(16, n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        if self.return_sequence:
            return self.module(x)
        else:
            return self.module(x[:, -1, :])


class SharPar(TSFEDL_BaseModule):
    """
    A 1-layer CNN + LSTM model. Originally proposed for depression detection.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=SharPar_Classifier(256, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Sharma, G., Parashar, A., & Joshi, A. M. (2021). DepHNN: a 
        novel hybrid neural network for electroencephalogram 
        (EEG)-based screening of depression. Biomedical signal 
        processing and control, 66, 102393.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = SharPar_Classifier(256, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(SharPar, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.convolutions = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=64, kernel_size=5, stride=1, bias=True, padding="same"),
            nn.ReLU()
        )

        self.lstm1 = nn.LSTM(input_size=64, hidden_size=32, batch_first=True)
        self.lstm2 = nn.LSTM(input_size=32, hidden_size=16, batch_first=True)

        self.fcc_module = nn.Sequential(
            nn.Linear(16, 16),
            nn.ReLU(),
            nn.Linear(16, 16),
            nn.ReLU()
        )

    def forward(self, x):
        x = self.convolutions(x)
        x = flip_indices_for_conv_to_lstm_reshape(x)
        x, _ = self.lstm1(x)
        x, _ = self.lstm2(x)
        x = self.fcc_module(x)

        if self.classifier is not None:
            x = self.classifier(x)
        return x

class DaiXiLi_Classifier(nn.Module):
    """
    Classifier of the DaiXiLi model.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        n_classes: int
            Number of classes to predict at the end of the network

        return_sequence: bool, defaults=True
            Parameter that controls wether the module returns the sequence or the prediction.

    Returns
    -------
    `LightningModule`
        A pyTorch Lightning Module instance.
    """
    def __init__(self, in_features, n_classes, return_sequence = False):
        super(DaiXiLi_Classifier, self).__init__()
        self.return_sequence = return_sequence
        self.module = nn.Sequential(
            nn.Linear(2048, n_classes),
            nn.Softmax()
        )

    def forward(self, x):
        return self.module(x)


class DaiXiLi(TSFEDL_BaseModule):
    """
    A 1-layer CNN + LSTM model. Originally proposed for depression detection.

    Parameters
    ----------
        in_features: int
            Number of features of the input tensors

        top_module: nn.Module, defaults=DaiXiLi_Classifier(256, 5)
            The optional  nn.Module to be used as additional top layers.

        loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            The loss function to use. It should accept two Tensors as inputs (predictions, targets) and return
            a Tensor with the loss.

        metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]]
            Dictionary with the name of the metric and a function to compute the metric from two tensors,
            prediction and true labels.

        optimizer:  torch.optim.Optimizer
            The pyTorch Optimizer to use. Note that this must be only the class type and not an instance of the class!!

        **kwargs: dict
            A dictionary with the parameters of the optimizer.

    Returns
    -------
        `LightningModule`
            A pyTorch Lightning Module instance.

    References
    ----------
        Sharma, G., Parashar, A., & Joshi, A. M. (2021). DepHNN: a 
        novel hybrid neural network for electroencephalogram 
        (EEG)-based screening of depression. Biomedical signal 
        processing and control, 66, 102393.
   """
    def __init__(self,
                 in_features: int,
                 top_module: Optional[nn.Module] = DaiXiLi_Classifier(256, 5),
                 loss: Callable[[torch.Tensor, torch.Tensor], torch.Tensor] = nn.CrossEntropyLoss(),
                 metrics: Dict[str, Callable[[torch.Tensor, torch.Tensor], torch.Tensor]] = None,
                 optimizer: torch.optim.Optimizer = torch.optim.Adam,
                 **kwargs
                 ):
        super(DaiXiLi, self).__init__(in_features, top_module, loss, metrics, optimizer, **kwargs)

        self.v2c1 = nn.Conv1d(in_channels=64, out_channels=32, kernel_size=5, stride=1, bias=True, padding="same")
        self.v2mp1 = nn.MaxPool1d(kernel_size=2, stride=2)
        self.v2d1 = nn.Linear(125, 16)

        self.v3c1 = nn.Conv1d(in_channels=in_features, out_channels=128, kernel_size=5, stride=1, bias=True, padding="same")
        self.v3mp1 = nn.MaxPool1d(kernel_size=2, stride=2)
        self.v3c2 = nn.Conv1d(in_channels=128, out_channels=64, kernel_size=5, stride=1, bias=True, padding="same")
        self.v3mp2 = nn.MaxPool1d(kernel_size=2, stride=2)
        self.v3c3 = nn.Conv1d(in_channels=64, out_channels=32, kernel_size=5, stride=1, bias=True, padding="same")
        self.v3mp3 = nn.MaxPool1d(kernel_size=2, stride=2)
        self.v3d1 = nn.Linear(125, 16)

        self.v1c1 = nn.Conv1d(in_channels=128, out_channels=64, kernel_size=5, stride=1, bias=True, padding="same")
        self.v1mp1 = nn.MaxPool1d(kernel_size=2, stride=2)
        self.v1d1 = nn.Linear(250, 16)

    def forward(self, x):
        x3 = self.v3c1(x)
        x3 = self.v3mp1(x3)

        x1 = self.v1c1(x3)
        x1 = self.v1mp1(x1)
        x1 = self.v1d1(x1)
        
        x3 = self.v3c2(x3)
        x3 = self.v3mp2(x3)

        x2 = self.v2c1(x3)
        x2 = self.v2mp1(x2)
        x2 = self.v2d1(x2)

        x3 = self.v3c3(x3)
        x3 = self.v3mp3(x3)
        x3 = self.v3d1(x3)

        x = torch.cat((x1, x2, x3), 1)
        x = x.view(x.size(0), -1)

        if self.classifier is not None:
            x = self.classifier(x)
        return x


import torch
import torch.nn as nn
import torch.nn.functional as F


class ConvBlockYiboGao(nn.Module):
    """
    Convolutional block of YiboGao's model

    Parameters
    ----------
    in_features : int
        Number of input features.
    nb_filter : int
        Number of filters for the convolution.
    kernel_size : int
        Size of the convolution kernel.
    """
    def __init__(self, in_features, nb_filter, kernel_size):
        super(ConvBlockYiboGao, self).__init__()
        self.kernel_size = kernel_size
        self.nb_filter = nb_filter
        self.in_features = in_features
        self.out_features = nb_filter

        self.module = nn.Sequential(
            nn.Conv1d(in_channels=in_features, out_channels=nb_filter, kernel_size=kernel_size, padding='same'),
            nn.BatchNorm1d(nb_filter),
            nn.ReLU()
        )

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    x : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):
        return self.module(x)


class AttentionBranchYiboGao(nn.Module):
    """
    Attention branch of YiboGao's model

    Parameters
    ----------
    in_features : int
        Number of input features.
    nb_filter : int
        Number of filters for the convolution.
    kernel_size : int
        Size of the convolution kernel.
    """

    def __init__(self, in_features, nb_filter, kernel_size):
        super(AttentionBranchYiboGao, self).__init__()
        self.nb_filter = nb_filter
        self.kernel_size = kernel_size
        self.in_features = in_features
        self.out_features = nb_filter

        self.convBlock1 = ConvBlockYiboGao(in_features, nb_filter, kernel_size)  # nb_filter is the number of features
        self.convBlock2 = nn.Sequential(
            nn.MaxPool1d(kernel_size=2),
            ConvBlockYiboGao(in_features=nb_filter, nb_filter=nb_filter, kernel_size=kernel_size),
            nn.Upsample(scale_factor=2)
        )
        self.convBlock3 = ConvBlockYiboGao(nb_filter, nb_filter, kernel_size)
        self.finalBlock = nn.Sequential(
            ConvBlockYiboGao(nb_filter, nb_filter, kernel_size),
            nn.Conv1d(in_channels=nb_filter, out_channels=nb_filter, kernel_size=1, padding='same'),
            nn.BatchNorm1d(nb_filter),
            nn.Sigmoid()
        )

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    x : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):
        x1 = self.convBlock1(x)
        x = self.convBlock2(x1)
        x2 = self.convBlock3(x)

        if(x1.size() != x2.size()):
            # For odd input sizes, the convolutions could produce differences of length 1 in sizes. Pad with zeroes
            x2 = F.pad(x2, pad=(0, 1))

        x = torch.add(x1, x2)

        return self.finalBlock(x)


class RTABlock(nn.Module):
    """
    Residual-based Temporal Attention (RTA) block.

    References
    ----------
        Gao, Y., Wang, H., & Liu, Z. (2021). An end-to-end atrial fibrillation detection by a novel residual-based
        temporal attention convolutional neural network with exponential nonlinearity loss.
        Knowledge-Based Systems, 212, 106589.

    Parameters
    ----------
    in_features : int
        Number of input features.
    nb_filter : int
        Number of filters for the convolution.
    kernel_size : int
        Size of the convolution kernel.
    """
    def __init__(self, in_features, nb_filter, kernel_size):
        super(RTABlock, self).__init__()
        self.nb_filter = nb_filter
        self.kernel_size = kernel_size
        self.in_features = in_features

        self.conv1 = ConvBlockYiboGao(in_features, nb_filter, kernel_size)
        self.conv2 = ConvBlockYiboGao(nb_filter, nb_filter, kernel_size)
        self.attention = AttentionBranchYiboGao(nb_filter, nb_filter, kernel_size)
        self.conv3 = ConvBlockYiboGao(nb_filter, nb_filter, kernel_size)

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    x : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):
        x1 = self.conv1(x)
        x2 = self.conv2(x1)
        attention_map = self.attention(x1)

        x = torch.multiply(x2, attention_map)
        x = torch.add(x, x1)

        return self.conv3(x)


class SqueezeAndExcitationModule(nn.Module):
    """
    Squeeze-and-Excitation Module.

    References
    ----------
     Squeeze-and-Excitation Networks, Jie Hu, Li Shen, Samuel Albanie, Gang Sun, Enhua Wu (arXiv:1709.01507v4)

    Parameters
    ---------
      in_features: int
        The number of input features (channels)

      dense_units: int
        The number units on each dense layer.

    Returns
    -------
    se: torch.Tensor
      Output tensor for the block.
    """
    def __init__(self, in_features: int, dense_units: int):
        super(SqueezeAndExcitationModule, self).__init__()
        self.dense_units = dense_units
        self.in_features = in_features

        self.fully_connected = nn.Sequential(
            nn.Linear(in_features, dense_units),
            nn.ReLU(),
            nn.Linear(dense_units, in_features),
            nn.Sigmoid()
        )

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    se : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):  # x shape is (N, C, L) N -> num_batch, C -> channels, L -> length of sequence
        # Global average pooling is just getting the average value of each channel
        se = torch.mean(x, dim=2)

        # reshape se to match x's shape
        se = se.view((se.size(0), 1, se.size(1)))  # change to (N, *, H_in) format
        se = self.fully_connected(se)
        se = se.view((se.size(0), se.size(2), 1))  # Go back to the (N, C, L) format to continue the convolutions.

        # element-wise multiplication between inputs and squeeze
        se = torch.multiply(x, se)
        return se


class DenseNetTransitionBlock(nn.Module):
    """
    Densenet Transition Block for CaiWenjuan model.

    Parameters
    ---------
      in_features: int
        The number of input features (channels)

      reduction: float
        Number between 0 and 1 representing the percentage reduction on the number
        of units.
    """
    def __init__(self, in_features, reduction):
        super(DenseNetTransitionBlock, self).__init__()
        self.out_features = int(in_features * reduction)
        self.module = nn.Sequential(
            nn.BatchNorm1d(num_features=in_features, eps=1.001e-5),
            nn.ReLU(),
            nn.Conv1d(in_channels=in_features, out_channels=int(in_features * reduction), kernel_size=1, bias=False),
            nn.AvgPool1d(kernel_size=2, stride=2)
        )

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    se : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):
        return self.module(x)


class DenseNetConvBlock(nn.Module):
    """
    Densenet convolution block.

    Parameters
    ---------
      in_features: int
        The number of input features (channels)

      growth_rate: int
        Growth rate of the number of units in the layers.
    """
    def __init__(self, in_features, growth_rate):
        super(DenseNetConvBlock, self).__init__()
        self.module = nn.Sequential(
            nn.BatchNorm1d(num_features=in_features, eps=1.001e-5),
            nn.ReLU(),
            nn.Conv1d(in_channels=in_features, out_channels=4 * growth_rate, kernel_size=1, bias=False),
            nn.BatchNorm1d(num_features=4 * growth_rate),
            nn.ReLU(),
            nn.Conv1d(in_channels=4 * growth_rate, out_channels=growth_rate, kernel_size=3, padding='same', bias=False)
        )

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    se : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):
        x1 = self.module(x)
        return torch.cat((x1, x), dim=1)


class DenseNetDenseBlock(nn.Module):
    """
    Densenet dense block.

    Parameters
    ---------
    in_features: int
        The number of input features (channels)

    layers : int
        Number of layers of the block.

    growth_rate: int
        Growth rate of the number of units in the layers.
    """
    def __init__(self, in_features, layers, growth_rate):
        super(DenseNetDenseBlock, self).__init__()
        self.output_features = in_features + layers * growth_rate
        self.module = nn.ModuleList(
            [DenseNetConvBlock(in_features=in_features + i * growth_rate, growth_rate=growth_rate)
             for i in range(layers)]
        )

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    se : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):
        xs = x
        for layer in self.module:
            xs = layer(xs)
        return xs


class SpatialAttentionBlockZhangJin(nn.Module):
    """
    Spatial Attention module of ZhangJin's model.

    Parameters
    ---------
    in_features: int
        The number of input features (channels).

    decrease_ratio: int
        Decrease rate of the number of units in the layers.
    """
    def __init__(self, in_features, decrease_ratio):
        super(SpatialAttentionBlockZhangJin, self).__init__()
        self.shared_dense = nn.Sequential(
            nn.Linear(in_features=in_features, out_features=in_features // decrease_ratio),
            nn.ReLU(),
            nn.Linear(in_features=in_features // decrease_ratio, out_features=in_features),
            nn.ReLU()
        )

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    se : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):
        # 1º Global average pooling for each feature
        x1 = torch.mean(x, dim=2)
        x1 = self.shared_dense(x1)
        x1 = x1.view((x1.size(0), x1.size(1), 1))

        # 2º Global max pooling
        x2 = torch.max(x, dim=2).values
        x2 = self.shared_dense(x2)
        x2 = x2.view((x2.size(0), x2.size(1), 1))

        result = F.sigmoid(torch.add(x1, x2))
        return result


class TemporalAttentionBlockZhangJin(nn.Module):
    """
    Temporal attention module of ZhangJin's Model.
    """
    def __init__(self):
        super(TemporalAttentionBlockZhangJin, self).__init__()
        self.conv1 = nn.Conv1d(in_channels=2, out_channels=1, kernel_size=7, padding='same')

    """
    Forward function of the model. This function returns the operation of the
    neural network over data.

    Parameters
    ----------
    x : array-like
        Data to operate with.

    Returns
    -------
    se : torch.Tensor
        Result of the operation with the neural network.
    """
    def forward(self, x):
        # Global max pooling for each TIMESTEP
        x1 = torch.max(x, dim=1).values
        x1 = x1.view((x1.size(0), 1, x1.size(1)))

        x2 = torch.mean(x, dim=1)
        x2 = x2.view((x2.size(0), 1, x2.size(1)))

        x = torch.cat((x1, x2), dim=1)
        x = self.conv1(x)
        x = F.sigmoid(x)

        return x


#
# def temporal_attention_block_ZhangJin(x):
#     """
#     Temporal attention module of ZhangJin's Model.
#     """
#     # Temporal attention module
#     x1 = layers.GlobalMaxPool1D(data_format='channels_first')(x)
#     x1 = layers.Reshape(target_shape=(x1.shape[1], 1))(x1)
#
#     x2 = layers.GlobalAveragePooling1D(data_format='channels_first')(x)
#     x2 = layers.Reshape(target_shape=(x2.shape[1], 1))(x2)
#
#     x = layers.Concatenate()([x1, x2])
#     x = layers.Conv1D(filters=1, kernel_size=7, padding="same")(x)
#     x = layers.Activation(activation=sigmoid)(x)
#     return x



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
import torch
from torch import nn
from torch.nn import functional as F

class TimeDistributed(nn.Module):
    """
    TimeDistributed module implementation.
    """
    def __init__(self, module):
        super(TimeDistributed, self).__init__()
        self.module = module

    def forward(self, x):
        if len(x.size()) <= 2:
            return self.module(x)
        t, n = x.size(0), x.size(1)
        # merge batch and seq dimensions
        x_reshape = x.contiguous().view(t * n, *x.size()[2:]).to(torch.float32)
        y = self.module(x_reshape)
        # We have to reshape Y
        y = y.contiguous().view(t, n, *y.size()[1:]).to(torch.float32)
        return y


def flip_indices_for_conv_to_lstm(x: torch.Tensor) -> torch.Tensor:
    """
    Changes the (N, C, L) dimension to (N, L, C). This is due to features in PyTorch's LSTMs are expected on the last dim.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor.

    Returns
    -------
    x : torch.Tensor
        Output tensor.
    """
    return x.view(x.size(0), x.size(2), x.size(1))

def flip_indices_for_conv_to_lstm_reshape(x: torch.Tensor) -> torch.Tensor:
    """
    Changes the (N, C, L) dimension to (N, L, C). This is due to features in PyTorch's LSTMs are expected on the last dim.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor.

    Returns
    -------
    x : torch.Tensor
        Output tensor.
    """
    return x.reshape(x.size(0), x.size(2), x.size(1))


def check_inputs(include_top, weights, input_tensor, input_shape, classes, classifier_activation):
    """
    Auxiliar function for checking the input parameters of the models.

    Parameters
    ----------
    include_top : bool
        Boolean value to control if the classification module should be placed in the model.
    weights : str
        Route to the saved weight of the model.
    input_tensor : keras.Tensor
        Input tensor of the model.
    input_shape : tuple
        Tuple with the input shape of the model.
    classes : int
        Number of classes to predict with the model.
    classifier_activation : str
        "softmax" or None

    Returns
    -------
    inp : Keras.Tensor
        Input tensor.
    """
    if include_top:
        if not isinstance(classes, int):
            raise ValueError("'classes' must be an int value.")
        act = keras.activations.get(classifier_activation)
        if act not in {keras.activations.get('softmax'), keras.activations.get(None)}:
            raise ValueError("'classifier_activation' must be 'softmax' or None.")

    if weights is not None and not tf.io.gfile.exists(weights):
        raise ValueError("'weights' path does not exists: ", weights)

    # Determine input
    if input_tensor is None:
        if input_shape is not None:
            inp = layers.Input(shape=input_shape)
        else:
            raise ValueError("One of input_tensor or input_shape should not be None.")
    else:
        inp = input_tensor

    return inp


def full_convolution(x, filters, kernel_size, **kwargs):
    """
    It performs a Full convolution operation on the given keras Tensor.

    Parameters
    ----------
    x : Keras.Tensor
        Input tensor of the full convolution.
    filters : int
        Number of filters of the full convolution.
    kernel_size : int
        Kernel size of the convolution.
    kwargs : dict
        Rest of the arguments, optional.

    Returns
    -------
    x : Keras.Tensor
        Output tensor.
    """
    # Do a full convolution. Return a keras Tensor
    x = layers.ZeroPadding1D(padding=kernel_size - 1)(x)
    x = layers.Conv1D(filters=filters, kernel_size=kernel_size, **kwargs)(x)
    return x



from tensorflow.keras import backend as K
from tensorflow.keras import layers
from tensorflow.python.keras.activations import relu, sigmoid


def densenet_transition_block(x, reduction, name):
    """A transition block of densenet for 1D data.

    Parameters
    ----------
      x: input tensor.
      reduction: float
        Compression rate at transition layers.
      name: str
        Block label.

    Returns
    -------
    x : output tensor for the block.
    """
    bn_axis = 2
    x = layers.BatchNormalization(axis=2, epsilon=1.001e-5, name=name + '_bn')(x)
    x = layers.Activation('relu', name=name + '_relu')(x)
    x = layers.Conv1D(int(x.shape[bn_axis] * reduction), 1, use_bias=False, name=name + '_conv')(x)
    x = layers.AveragePooling1D(2, strides=2, name=name + '_pool')(x)
    return x


def densenet_conv_block(x, growth_rate, name):
    """A building block for a dense block from densenet for 1D data.

    Parameters
    ----------
      x: input tensor.
      growth_rate: float
        Growth rate at dense layers.
      name: str
        Block label.

    Returns
    -------
    x : Output tensor for the block.
    """
    bn_axis = 2  # 3 if backend.image_data_format() == 'channels_last' else 1
    x1 = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5, name=name + '_0_bn')(x)
    x1 = layers.Activation('relu', name=name + '_0_relu')(x1)
    x1 = layers.Conv1D(4 * growth_rate, 1, use_bias=False, name=name + '_1_conv')(x1)
    x1 = layers.BatchNormalization(axis=bn_axis, epsilon=1.001e-5, name=name + '_1_bn')(x1)
    x1 = layers.Activation('relu', name=name + '_1_relu')(x1)
    x1 = layers.Conv1D(growth_rate, 3, padding='same', use_bias=False, name=name + '_2_conv')(x1)
    x = layers.Concatenate(axis=bn_axis, name=name + '_concat')([x, x1])
    return x


def densenet_dense_block(x, blocks, growth_rate, name):
    """A dense block of densenet for 1D data.

    Parameters
    ----------
      x: input tensor.
      blocks: int
        The number of building blocks.
      name: str
        Block label.

    Returns
    -------
    x : Output tensor for the block.
    """
    for i in range(blocks):
        x = densenet_conv_block(x, growth_rate, name=name + '_block' + str(i + 1))
    return x


def squeeze_excitation_module(x, dense_units):
    """
    Squeeze-and-Excitation Module.

    References
    ----------
     Squeeze-and-Excitation Networks, Jie Hu, Li Shen, Samuel Albanie, Gang Sun, Enhua Wu (arXiv:1709.01507v4)

    Arguments
    ---------
      x: keras.Tensor
        The input tensor.

      dense_units: int
        The number units on each dense layer.

    Returns
    -------
    se: Keras.Tensor
      Output tensor for the block.
    """
    se = layers.GlobalAveragePooling1D()(x)
    # Reshape the output to mach the number of dimensions of the input shape: (None, 1, Channels)
    se = layers.Reshape(target_shape=(1, x.shape[2]))(se)
    se = layers.Dense(units=dense_units, activation=relu)(se)  # Fully-Connected 1
    se = layers.Dense(units=x.shape[2], activation=sigmoid)(se)  # Fully-Connected 2

    # Perform the element-wise multiplication between the inputs and the Squeeze
    se = layers.multiply([x, se])
    return se


def conv_block_YiboGao(in_x, nb_filter, kernel_size):
    """
    Convolutional block of YiboGao's model.

    Parameters
    ----------
    in_x : keras.Tensor
        Input tensor of the convolution bock.
    nb_filter : int
        Number of filerts for the convolution.
    kernel_size : int
        Kernel size of the convolution.

    Returns
    -------
    x : keras.Tensor
        Output tensor of the block.
    """
    x = layers.Conv1D(filters=nb_filter, kernel_size=kernel_size, padding='same')(in_x)
    x = layers.BatchNormalization()(x)
    x = layers.Activation(activation=relu)(x)
    return x


def attention_branch_YiboGao(in_x, nb_filter, kernel_size):
    """
    Attention bronch of YiboGao's model.

    Parameters
    ----------
    in_x : keras.Tensor
        Input tensor.
    nb_filter : int
        Number of filerts for the convolutional YiboGao block.
    kernel_size : int
        Kernel size for the convolutional block.

    Returns
    -------
    x : keras.Tensor
        Output tensor of the block.
    """
    x1 = conv_block_YiboGao(in_x, nb_filter, kernel_size)

    x = layers.MaxPooling1D(pool_size=2)(x1)
    x = conv_block_YiboGao(x, nb_filter, kernel_size)
    x = layers.UpSampling1D(size=2)(x)

    x2 = conv_block_YiboGao(x, nb_filter, kernel_size)

    if (K.int_shape(x1) != K.int_shape(x2)):
        x2 = layers.ZeroPadding1D(padding=1)(x2)
        x2 = layers.Cropping1D((1, 0))(x2)

    x = layers.add([x1, x2])

    x = conv_block_YiboGao(x, nb_filter, kernel_size)

    x = layers.Conv1D(filters=nb_filter, kernel_size=1, padding='same')(x)
    x = layers.BatchNormalization()(x)
    x = layers.Activation(activation=sigmoid)(x)

    return x


def RTA_block(in_x, nb_filter, kernel_size):
    """
    Residual-based Temporal Attention (RTA) block.

    References
    ----------
        Gao, Y., Wang, H., & Liu, Z. (2021). An end-to-end atrial fibrillation detection by a novel residual-based
        temporal attention convolutional neural network with exponential nonlinearity loss.
        Knowledge-Based Systems, 212, 106589.

    Parameters
    ----------
    in_x : keras.Tensor
        Input tensor.
    nb_filter : int
        Number of filerts for the convolutional YiboGao block.
    kernel_size : int
        Kernel size for the convolutional block.

    Returns
    -------
    out : keras.Tensor
        Output tensor of the block.
    """
    x1 = conv_block_YiboGao(in_x, nb_filter, kernel_size)
    x2 = conv_block_YiboGao(x1, nb_filter, kernel_size)

    attention_map = attention_branch_YiboGao(x1, nb_filter, kernel_size)

    x = layers.multiply([x2, attention_map])
    x = layers.add([x, x1])

    out = conv_block_YiboGao(x, nb_filter, kernel_size)

    return out


def spatial_attention_block_ZhangJin(decrease_ratio, x):
    """
    Spatial attention module of ZhangJin's model

    Parameters
    ----------
    x : keras.Tensor
        Input tensor.
    decrease_ratio : int
        Decrease ratio of the number of units in the neural network.

    Returns
    -------
    x : keras.Tensor
        Output tensor of the block.
    """
    shared_dense1 = layers.Dense(units=x.shape[2] // decrease_ratio, activation=relu)
    shared_dense2 = layers.Dense(units=x.shape[2], activation=relu)
    x1 = layers.GlobalAveragePooling1D()(x)
    x1 = shared_dense1(x1)
    x1 = shared_dense2(x1)
    x1 = layers.Reshape(target_shape=(1, x.shape[2]))(x1)

    x2 = layers.GlobalMaxPool1D()(x)
    x2 = shared_dense1(x2)
    x2 = shared_dense2(x2)
    x2 = layers.Reshape(target_shape=(1, x.shape[2]))(x2)

    x = layers.add([x1, x2])
    x = layers.Activation(activation=sigmoid)(x)
    return x


def temporal_attention_block_ZhangJin(x):
    """
    Temporal attention module of ZhangJin's Model.

    Parameters
    ----------
    x : keras.Tensor
        Input tensor.

    Returns
    -------
    x : keras.Tensor
        Output tensor of the block.
    """
    # Temporal attention module
    x1 = layers.GlobalMaxPool1D(data_format='channels_first')(x)
    x1 = layers.Reshape(target_shape=(x1.shape[1], 1))(x1)

    x2 = layers.GlobalAveragePooling1D(data_format='channels_first')(x)
    x2 = layers.Reshape(target_shape=(x2.shape[1], 1))(x2)

    x = layers.Concatenate()([x1, x2])
    x = layers.Conv1D(filters=1, kernel_size=7, padding="same")(x)
    x = layers.Activation(activation=sigmoid)(x)
    return x



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 14 17:16:34 2022
Python TSFEDL Setup.
"""

from setuptools import setup, find_packages

from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    #  Project name.
    #  $ pip install pyDML
    name='TSFEDL',

    # Version
    version='1.0.7.5',

    # Description
    description='Time Series Spatio-Temporal Feature Extraction using Deep Learning',

    # Long description (README)
    long_description=long_description,

    # URL
    url='https://github.com/ari-dasci/S-TSFE-DL',

    # Author
    author='Ignacio Aguilera Martos, Ángel Miguel García Vico, Julian Luengo, Francisco Herrera',

    # Author email
    author_email='nacheteam@ugr.es',

    # Keywords
    keywords=['Time series',
              'Feature extraction',
              'Deep learning',
              'recurrent',
              'cnn'],

    # Packages
    packages=find_packages(exclude=['docker', 'docs', 'test', 'examples']),

    # Test suite
    test_suite='test',

    # Requeriments
    install_requires=['scikit-learn', 'wfdb', 'obspy', 'tensorflow', 'keras', 'pytorch-lightning', 'torchmetrics'],

    long_description_content_type='text/markdown',
    python_requires='>=3.8'
)
